static const char _rcsid_[]="$Id: xio-openssl.c,v 1.7 2003/02/23 07:15:25 gerhard Exp $";
/* Copyright Gerhard Rieger 2002 */
/* Published under the GNU General Public License V.2, see file COPYING */

/* this file contains the implementation of the openssl addresses */

#include "xiosysincludes.h"
#if WITH_OPENSSL	/* make this address configure dependend */
#include "xioopen.h"

#include "xio-fd.h"
#include "xio-ipapp.h"
#include "xio-openssl.h"

/* the openssl library requires a file descriptor for external communications.
   so our best effort is to provide any possible kind of un*x file descriptor 
   (not only tcp, but also pipes, stdin, files...)
   for tcp we want to provide support for socks and proxy.
   read and write functions must use the openssl crypt versions.
*/

/* Linux: "man 3 ssl" */

/* generate a simple openssl server for testing:
   1) generate a private key
   openssl genrsa -out server.key 1024
      enter fields...
   2) generate a self signed cert
   openssl req -new -key server.key -x509 -days 3653 -out server.crt
      enter fields...
   3) generate the pem file
   cat server.key server.crt >server.pem
   openssl s_server  (listens on 4433/tcp)
 */

/* static declaration of ssl's open function */
static int xioopen_openssl_connect(char *arg, int rw, xiofile_t *fd, unsigned groups,
			   int dummy1, int dummy2, void *dummyp1);


/* description record for ssl */
const struct addrdesc addr_openssl = {
   "openssl",	/* keyword for selecting this address type in xioopen calls
		   (canonical or main name) */
   3,		/* data flow directions this address supports on API layer:
		   1..read, 2..write, 3..both */
   xioopen_openssl_connect,	/* a function pointer used to "open" these addresses.*/
   GROUP_SOCKET|GROUP_SOCK_IP|GROUP_IP_TCP|GROUP_OPENSSL,	/* bitwise OR of address groups this address belongs to.
		   You might have to specify a new group in xioopts.h */
   0,		/* an integer passed to xioopen_openssl; makes it possible to
		   use the same xioopen_openssl function for slightly different
		   address types. */
   0,		/* like previous argument */
   NULL		/* like previous arguments, but pointer type.
		   No trailing comma or semicolon! */
   HELP(":<host>:<port>")	/* a text displayed from xio help function.
			   No trailing comma or semicolon!
			   only generates this text if WITH_HELP is != 0 */
} ;

const struct optdesc opt_openssl_cipherlist = { "openssl-cipherlist", "ciphers", OPT_OPENSSL_CIPHERLIST, GROUP_OPENSSL, PH_SPEC, TYPE_STRING, OFUNC_SPEC };
const struct optdesc opt_openssl_method     = { "openssl-method",     "method",  OPT_OPENSSL_METHOD,     GROUP_OPENSSL, PH_SPEC, TYPE_STRING, OFUNC_SPEC };
const struct optdesc opt_openssl_verify     = { "openssl-verify",     "verify",  OPT_OPENSSL_VERIFY,     GROUP_OPENSSL, PH_SPEC, TYPE_INT,    OFUNC_SPEC };

/* the open function */
static int
   xioopen_openssl_connect(char *arg,	/* the arguments in the address string */
		   int rw,	/* is the open meant for reading (0),
				   writing (1), or both (2) ? */
		   xiofile_t *xfd,	/* a xio file descriptor structure,
				   already allocated */
		   unsigned groups,	/* the matching address groups... */
		   int dummy1,	/* first transparent integer value from
				   addr_openssl */
		   int dummy2,	/* second transparent integer value from
				   addr_openssl */
		   void *dummyp1)	/* transparent pointer value from
					   addr_openssl */
{
   struct opt *opts;
   SSL_CTX* ctx;
   SSL *ssl;
   SSL_METHOD *method;
   X509 *server_cert;
   char *str;
   char *me_str = NULL;	/* method string */
   char *ci_str = NULL;	/* cipher string */
   bool opt_ver = false;	/* dont verify peer certificate */
   /*0 int vermode;*/		/* peer certificate verification mode */
   char buff[2048];	/* hold server certificate */
   char error_string[120];
   unsigned long err;
   int result, ret, errint;

   /* parse the required arguments in arg */
#if 0
   if (*arg == ':') {
#endif
      char *colon, *comma;

   if (colon = strchr(arg, ':')) {
      *colon++ = '\0';
   } else {
      Error1("xioopen_openssl_connect(\"%s\", ...): missing parameter(s)", arg);
      return -1;
   }

      if (comma = strchr(colon, ',')) {
	 *comma++ = '\0';
      }
      if (parseopts(comma, groups, &opts) < 0) {
	 return STAT_NORETRY;
      }
      retropt_string(opts, OPT_OPENSSL_METHOD, &me_str);
      retropt_string(opts, OPT_OPENSSL_CIPHERLIST, &ci_str);
      retropt_bool(opts, OPT_OPENSSL_VERIFY, &opt_ver);

      if (result =
	  _xioopen_ip4app_connect(arg, colon, xfd,
				  SOCK_STREAM, IPPROTO_TCP, "tcp", opts) < 0) {
	 return result;
      }

      /*! isn't this too early? */
   if ((result = _xio_openlate(&xfd->stream, opts)) < 0) {
      return result;
   }

#if 0
   } else if (*arg == ',') {
      /* need transport option */

   } else {
      Error("openssl: missing target specification");
      return -1;
   }
#endif

   xfd->common.addr  = &addr_openssl;
   xfd->stream.dtype = DATA_OPENSSL;

   /* invoke parseopts() for options in arg string */

   /* try to contact or create the addressed resource */

   /*! actions_to_seed_PRNG();*/

   if (me_str != 0) {
      if (!strcasecmp(me_str, "SSLv2") || !strcasecmp(me_str, "SSL2")) {
	 method = sycSSLv2_client_method();
      } else if (!strcasecmp(me_str, "SSLv3") || !strcasecmp(me_str, "SSL3")) {
	 method = sycSSLv3_client_method();
      } else if (!strcasecmp(me_str, "SSLv23") || !strcasecmp(me_str, "SSL23") ||
		 !strcasecmp(me_str, "SSL")) {
	 method = sycSSLv23_client_method();
      } else if (!strcasecmp(me_str, "TLSv1") || !strcasecmp(me_str, "TLS1") ||
		 !strcasecmp(me_str, "TLS")) {
	 method = sycTLSv1_client_method();
      } else {
	 Error1("openssl-method=\"%s\": unknown method", me_str);
	 method = sycSSLv23_client_method()/*!*/;
      }
   } else {
      method = sycSSLv23_client_method()/*!*/;
   }

   /* OpenSSL preparation */
   sycSSL_library_init();

   sycSSL_load_error_strings();
      
   if ((ctx = sycSSL_CTX_new(method)) == NULL) {
      if (ERR_peek_error() == 0) Error("SSL_CTX_new()");
      while (err = ERR_get_error()) {
	 Error1("SSL_CTX_new(): %s", ERR_error_string(err, NULL));
      }

      /*ERR_clear_error;*/
      return STAT_NORETRY;
   }

   /* create a SSL object */
   if ((ssl = sycSSL_new(ctx)) == NULL) {
      if (ERR_peek_error() == 0) Error("SSL_new() failed");
      while (err = ERR_get_error()) {
	 Error1("SSL_new(): %s", ERR_error_string(err, NULL));
      }
      /*Error("SSL_new()");*/
      return STAT_NORETRY;
   }
   xfd->stream.para.openssl.ssl = ssl;

   /* set pre ssl-connect options */
   /* SSL_CIPHERS */
   if (ci_str != NULL) {
      if (sycSSL_set_cipher_list(ssl, ci_str) <= 0) {
	 if (ERR_peek_error() == 0)
	    Error1("SSL_set_cipher_list(, \"%s\") failed", ci_str);
	 while (err = ERR_get_error()) {
	    Error2("SSL_set_cipher_list(, \"%s\"): %s",
		   ci_str, ERR_error_string(err, NULL));
	 }
	 /*Error("SSL_new()");*/
	 return STAT_NORETRY;
      }
   }

#if 0
   /* set required authentication of server */
   vermode = SSL_VERIFY_PEER;
   if (SSL_set_verify(ssl, vermode, xio_openssl_verify_callback)
       /*!*/;
#endif

   /* assign a network connection to the SSL object */
   if (sycSSL_set_fd(ssl, xfd->stream.fd) <= 0) {
      if (ERR_peek_error() == 0) Error("SSL_set_fd() failed");
      while (err = ERR_get_error()) {
	 Error2("SSL_set_fd(, %d): %s",
		xfd->stream.fd, ERR_error_string(err, NULL));
      }
   }

   /* connect via SSL by performing handshake */
   if ((ret = sycSSL_connect(ssl)) <= 0) {
      /*if (ERR_peek_error() == 0) Error("SSL_connect() failed");*/
      errint = SSL_get_error(ssl, ret);
      switch (errint) {
      case SSL_ERROR_NONE:
	 Error("ok"); break;
      case SSL_ERROR_ZERO_RETURN:
	 Error("connection closed (wrong version number?)"); break;
      case SSL_ERROR_WANT_READ: case SSL_ERROR_WANT_WRITE:
      case SSL_ERROR_WANT_CONNECT:
      case SSL_ERROR_WANT_X509_LOOKUP:
	 Error("nonblocking operation did not complete"); break;	/*!*/
      case SSL_ERROR_SYSCALL:
	 if (ERR_peek_error() == 0) {
	    if (ret == 0) {
	       Error("SSL_connect(): socket closed by peer");
	    } else if (ret == -1) {
	       Error1("SSL_connect(): %s", strerror(errno));
	    }
	 } else {
	    Error("I/O error");	/*!*/
	    while (err = ERR_get_error()) {
	       ERR_error_string_n(err, error_string, sizeof(error_string));
	       Error4("SSL_connect(): %s / %s / %s / %s", error_string,
		      ERR_lib_error_string(err), ERR_func_error_string(err),
		      ERR_reason_error_string(err));
	    }
	 }
	 break;
      case SSL_ERROR_SSL:
	 ERR_print_errors_fp(stderr);	/*!*/
	 break;
      default:
	 Error("unknown error");
      }
      return STAT_NORETRY;
   }

   /* SSL_CTX_add_extra_chain_cert
      SSL_get_verify_result
   */
   Notice1("SSL connection using %s", SSL_get_cipher(ssl));
   if ((server_cert = SSL_get_peer_certificate(ssl)) == NULL) {
      Notice("no peer certificate");
   } else {
      Debug("Server certificate:");
      if ((str = X509_NAME_oneline(X509_get_subject_name(server_cert), buff, sizeof(buff))) != NULL)
	 Debug1("\tsubject: %s", str);  /*free (str); SIGSEGV*/
      if ((str = X509_NAME_oneline(X509_get_issuer_name(server_cert), buff, sizeof(buff))) != NULL)
	 Debug1("\tissuer: %s", str);  /*free (str); SIGSEGV*/
   }

   if (opt_ver) {
      if (server_cert == NULL) {
	 Error("no peer certificate");
	 return STAT_NORETRY;
      } else {
	 Error("no check of certificate implemented");
	 return STAT_NORETRY;
      }
   } else {
      Warn("no check of certificate");
   }

#if LATER
   if ((verify_result = SSL_get_verify_result(ssl)) == X509_V_OK) {
      Info("accepted server certificate");
   } else {
     Error1("rejected server certificate with error %ld", verify_result);
   }
#endif
   X509_free(server_cert);

   /* fill in the fd structure */
   return STAT_OK;
}


#endif /* WITH_OPENSSL */
